/********************************************************************************
 *
 *  urbanVictory engine
 *  Copyright (C) 2003, 2004 Adam Cubitt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *										
 ********************************************************************************/



/**
 *	\file	matrix.h
 *	\author	Adam Cubitt
 *	\brief	defines a matrix class
 */

#ifndef __matrix4x4__
#define __matrix4x4__


#include <cmath>

#include "utilities.h"
#include "vector.h"
#include "matrix.h"


#pragma warning(disable:4786)

/// 4x4 matrix class
/**
 *	templated to make conversion from one type to another easier
 */
template<typename T>
class matrix4x4
{
public:
	/// set matrix to identity
	matrix4x4<T>() { setIdentity(); }

	/// constructor with assignement from array
	matrix4x4<T>(const T init[16]) { for(int i=0;i<16;++i) data[i]=init[i]; }
	
	/// multiplication of matricies
	matrix4x4<T> operator*(const matrix4x4<T>& b);
	
	/// multiplication with vector
	vector3f operator*(const vector3f& b);
	
	/// assignment for matrices
	matrix4x4<T> operator=(const matrix4x4<T>& b);

	/// equality operator for matricies
	bool operator==(const matrix4x4<T>& b) const { for(int i=0;i<16;i++) if(data[i]!=b.data[i]) return false;	return true; }

	/// allows matrix elements to be accessed like an array
	T& operator[](const unsigned int i) { ASSERT(i<16,"matrix index out of range"); return data[i]; }

	/// allows matrix elements to be accessed like an array
	T operator[](const unsigned int i) const { ASSERT(i<16,"matrix index out of range"); return data[i]; }


	/// sets the matrix to the identity matrix;
	void setIdentity();

	/// returns the transpose of the matrix
	matrix4x4<T> getTranspose() const;
	
	/// returns the inverse of the matrix assuming no scaling
	matrix4x4<T> getInverseDirty() const;

	/// get the rotation as an euler rotation
	v3f getAsEulerRad() const;

	/// get the rotation as an euler rotation
	v3f getAsEulerDeg() const;

	/// set the matrix to the euler rotation specified
	void setFromEulerRad(const v3f& e);

	/// set the matrix to the euler rotation specified
	void setFromEulerDeg(const v3f& e);

	/// allows output of matrix data
	friend std::ostream &operator << (std::ostream& os, const matrix4x4<T>& m_);

	/// The matrix data
	T data[16];
};

typedef matrix4x4<float> matrix;


/**
*	\author Robert Bateman
*/
template<typename T>
matrix4x4<T> matrix4x4<T>::operator*(const matrix4x4<T>& b)
{
	matrix4x4<T> c;

	c.data[0] = data[0]*b.data[0];
	c.data[1] = data[0]*b.data[1];
	c.data[2] = data[0]*b.data[2];
	c.data[3] = data[0]*b.data[3];

	c.data[0] += data[1]*b.data[4];
	c.data[1] += data[1]*b.data[5];
	c.data[2] += data[1]*b.data[6];
	c.data[3] += data[1]*b.data[7];

	c.data[0] += data[2]*b.data[8];
	c.data[1] += data[2]*b.data[9];
	c.data[2] += data[2]*b.data[10];
	c.data[3] += data[2]*b.data[11];

	c.data[0] += data[3]*b.data[12];
	c.data[1] += data[3]*b.data[13];
	c.data[2] += data[3]*b.data[14];
	c.data[3] += data[3]*b.data[15];

	c.data[4] = data[4]*b.data[0];
	c.data[5] = data[4]*b.data[1];
	c.data[6] = data[4]*b.data[2];
	c.data[7] = data[4]*b.data[3];

	c.data[4] += data[5]*b.data[4];
	c.data[5] += data[5]*b.data[5];
	c.data[6] += data[5]*b.data[6];
	c.data[7] += data[5]*b.data[7];

	c.data[4] += data[6]*b.data[8];
	c.data[5] += data[6]*b.data[9];
	c.data[6] += data[6]*b.data[10];
	c.data[7] += data[6]*b.data[11];

	c.data[4] += data[7]*b.data[12];
	c.data[5] += data[7]*b.data[13];
	c.data[6] += data[7]*b.data[14];
	c.data[7] += data[7]*b.data[15];

	c.data[8] = data[8]*b.data[0];
	c.data[9] = data[8]*b.data[1];
	c.data[10] = data[8]*b.data[2];
	c.data[11] = data[8]*b.data[3];

	c.data[8] += data[9]*b.data[4];  
	c.data[9] += data[9]*b.data[5];
	c.data[10] += data[9]*b.data[6];
	c.data[11] += data[9]*b.data[7];

	c.data[8] += data[10]*b.data[8];  
	c.data[9] += data[10]*b.data[9]; 
	c.data[10] += data[10]*b.data[10];
	c.data[11] += data[10]*b.data[11];

	c.data[8] += data[11]*b.data[12];
	c.data[9] += data[11]*b.data[13]; 
	c.data[10] += data[11]*b.data[14];
	c.data[11] += data[11]*b.data[15];

	c.data[12] = data[12]*b.data[0];
	c.data[13] = data[12]*b.data[1];
	c.data[14] = data[12]*b.data[2];
	c.data[15] = data[12]*b.data[3];

	c.data[12] += data[13]*b.data[4];
	c.data[13] += data[13]*b.data[5];
	c.data[14] += data[13]*b.data[6];
	c.data[15] += data[13]*b.data[7];

	c.data[12] += data[14]*b.data[8];
	c.data[13] += data[14]*b.data[9];
	c.data[14] += data[14]*b.data[10];
	c.data[15] += data[14]*b.data[11];

	c.data[12] += data[15]*b.data[12];
	c.data[13] += data[15]*b.data[13];
	c.data[14] += data[15]*b.data[14];
	c.data[15] += data[15]*b.data[15];

	return c;
	
}

template<typename T>
matrix4x4<T> matrix4x4<T>::operator=(const matrix4x4<T>& b)
{
	data[0] = b.data[0];
	data[1] = b.data[1];
	data[2] = b.data[2];
	data[3] = b.data[3];

	data[4] = b.data[4];
	data[5] = b.data[5];
	data[6] = b.data[6];
	data[7] = b.data[7];

	data[8] = b.data[8];
	data[9] = b.data[9];
	data[10] = b.data[10];
	data[11] = b.data[11];

	data[12] = b.data[12];
	data[13] = b.data[13];
	data[14] = b.data[14];
	data[15] = b.data[15];

	return *this;
}

/*
// matrix*vector
template<typename T>
vector3f matrix4x4<T>::operator*(const vector3f& b)
{
	vector3f c;

	c.x = data[0]*b.x + data[1]*b.y + data[2]*b.z + data[3];
	c.y = data[4]*b.x + data[5]*b.y + data[6]*b.z + data[7];
	c.z = data[8]*b.x + data[9]*b.y + data[10]*b.z + data[11];

	return c;
}
*/

template<typename T>
vector3f matrix4x4<T>::operator*(const vector3f& b)
{
	vector3f c;

	c.x = data[0]*b.x + data[4]*b.y + data[8]*b.z + data[12];
	c.y = data[1]*b.x + data[5]*b.y + data[9]*b.z + data[13];
	c.z = data[2]*b.x + data[6]*b.y + data[10]*b.z + data[14];

	return c;
}

template<typename T>
void matrix4x4<T>::setIdentity()
{
	data[0] = 1;
	data[1] = 0;
	data[2] = 0;
	data[3] = 0;

	data[4] = 0;
	data[5] = 1;
	data[6] = 0;
	data[7] = 0;

	data[8] = 0;
	data[9] = 0;
	data[10] = 1;
	data[11] = 0;

	data[12] = 0;
	data[13] = 0;
	data[14] = 0;
	data[15] = 1;
}



template<typename T>
matrix4x4<T> matrix4x4<T>::getTranspose() const
{
	matrix4x4<T> result;
	
	result.data[0] = data[0];
	result.data[1] = data[4];
	result.data[2] = data[8];
	result.data[3] = data[12];
	
	result.data[4] = data[1];
	result.data[5] = data[5];
	result.data[6] = data[9];
	result.data[7] = data[13];
	
	result.data[8] = data[2];
	result.data[9] = data[6];
	result.data[10] = data[10];
	result.data[11] = data[14];
	
	result.data[12] = data[3];
	result.data[13] = data[7];
	result.data[14] = data[11];
	result.data[15] = data[15];
	
	return result;
}


/**
*	\author Robert Bateman
*/
/*
template<typename T>
matrix4x4<T> matrix4x4<T>::getInverseDirty() const
{
	matrix4x4<T> result;
	const T *ptr = data;

	// The rotational part of the matrix is simply the transpose of the
	// original matrix.
	result.data[0]   = *(  ptr);
	result.data[4]   = *(++ptr);
	result.data[8]   = *(++ptr);

	T Tx		 = *(++ptr);

	result.data[1]   = *(++ptr);
	result.data[5]   = *(++ptr);
	result.data[9]   = *(++ptr);

	T Ty		 = *(++ptr);

	result.data[2]   = *(++ptr);
	result.data[6]   = *(++ptr);
	result.data[10]  = *(++ptr);

	T Tz	     = *(++ptr);

	ptr = data;

	// result = -(Tm * Rm) to get the translation part of the inverse
	result.data[3]  = - *(  ptr) * Tx;
	result.data[7]  = - *(++ptr) * Tx;
	result.data[11] = - *(++ptr) * Tx;
	++ptr;

	result.data[3]  -=  *(++ptr) * Ty;
	result.data[7]  -=  *(++ptr) * Ty;
	result.data[11] -=  *(++ptr) * Ty;
	++ptr;

	result.data[3]  -=  *(++ptr) * Tz;
	result.data[7]  -=  *(++ptr) * Tz;
	result.data[11] -=  *(++ptr) * Tz;

	result.data[12] = 0;
	result.data[13] = 0;
	result.data[14] = 0;
	result.data[15] = 1;

	return result;
}
*/


template<typename T>
matrix4x4<T> matrix4x4<T>::getInverseDirty() const
{
	matrix4x4<T> result;
	const T *ptr = data;

	// The rotational part of the matrix is simply the transpose of the
	// original matrix.
	result.data[0]   = *(  ptr);
	result.data[4]   = *(++ptr);
	result.data[8]   = *(++ptr);

	T Tx		 = *(++ptr);

	result.data[1]   = *(++ptr);
	result.data[5]   = *(++ptr);
	result.data[9]   = *(++ptr);

	T Ty		 = *(++ptr);

	result.data[2]   = *(++ptr);
	result.data[6]   = *(++ptr);
	result.data[10]  = *(++ptr);

	T Tz	     = *(++ptr);

	ptr = data;

	// result = -(Tm * Rm) to get the translation part of the inverse
	result.data[12]  = - *(  ptr) * Tx;
	result.data[13]  = - *(++ptr) * Tx;
	result.data[14] = - *(++ptr) * Tx;
	++ptr;

	result.data[12]  -=  *(++ptr) * Ty;
	result.data[13]  -=  *(++ptr) * Ty;
	result.data[14] -=  *(++ptr) * Ty;
	++ptr;

	result.data[12]  -=  *(++ptr) * Tz;
	result.data[13]  -=  *(++ptr) * Tz;
	result.data[14] -=  *(++ptr) * Tz;

	result.data[3] = 0;
	result.data[5] = 0;
	result.data[7] = 0;

	result.data[15] = 1;

	return result;
}



/// \biblio http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToEuler/
template <typename T>
v3f matrix4x4<T>::getAsEulerRad() const
{
	v3f e;
	// Assuming the angles are in radians.
	if (data[1] > 1.0f - episilon) { // singularity at north pole
		e.yaw = atan2(data[8], data[10]);
		e.pitch = PI/2;
		e.roll = 0;
		return e;
	}
	if (data[1] < -1.0f + episilon) { // singularity at south pole
		e.yaw = atan2(data[8], data[10]);
		e.pitch = -PI/2;
		e.roll = 0;
		return e;
	}
	e.yaw = atan2(-data[2], data[0]);
	e.roll = atan2(-data[9], data[5]);
	e.pitch = asin(data[1]);
	return e;
}

template <typename T>
v3f matrix4x4<T>::getAsEulerDeg() const
{
	v3f e = getAsEulerRad();
	return v3f(	TO_DEG(e.pitch),
				TO_DEG(e.yaw),
				TO_DEG(e.roll)	);
}


/// \biblio Essential Mathematics For Computer Graphics Fast P.108
/*
template<typename T>
void matrix4x4<T>::setFromEulerRad(const v3f& e)
{
	T cp = cos(e.pitch);
	T sp = sin(e.pitch);
	T cy = cos(e.yaw);
	T sy = sin(e.yaw);
	T cr = cos(e.roll);
	T sr = sin(e.roll);

	data[0] = cy*cr+sy*sp*sr;
	data[1] = cp*sr;
	data[2] = -sy*cr+cy*sp*sr;
	data[3] = 0;
	data[4] = -cy*sr+sy*sp*cr;
	data[5] = cp*cr;
	data[6] = sy*sr+cy*sp*cr;
	data[7] = 0;
	data[8] = sy*cp;
	data[9] = -sp;
	data[10] = cy*cp;
	data[11] = 0;
	data[12] = 0;
	data[13] = 0;
	data[14] = 0;
	data[15] = 1;
}
*/

template<typename T>
void matrix4x4<T>::setFromEulerRad(const v3f& e)
{
	T cp = cos(e.pitch);
	T sp = sin(e.pitch);
	T cy = cos(e.yaw);
	T sy = sin(e.yaw);
	T cr = cos(e.roll);
	T sr = sin(e.roll);

	data[0] = cy*cr+sy*sp*sr;
	data[4] = cp*sr;
	data[8] = -sy*cr+cy*sp*sr;
	data[12] = 0;
	data[1] = -cy*sr+sy*sp*cr;
	data[5] = cp*cr;
	data[9] = sy*sr+cy*sp*cr;
	data[13] = 0;
	data[2] = sy*cp;
	data[6] = -sp;
	data[10] = cy*cp;
	data[14] = 0;
	data[3] = 0;
	data[7] = 0;
	data[11] = 0;
	data[15] = 1;
}

template<typename T>
void matrix4x4<T>::setFromEulerDeg(const v3f& e)
{
	setFromEulerRad( v3f(	TO_RAD(e.pitch),
							TO_RAD(e.yaw),
							TO_RAD(e.roll)	));
}


// for outputting matricies
template<typename T>
std::ostream &operator << (std::ostream& os, const matrix4x4<T>& m_)
{
	os << "[[";
	for(int i=0;i<16;i++)
	{
		os << m_.data[i] << " ";
		if(i==3||i==7||i==11)
		{
			os << "]\n [";
		}
	}
	return os << "]]";
}



#endif

