/********************************************************************************
 *
 *  urbanVictory engine
 *  Copyright (C) 2003, 2004 Adam Cubitt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *										
 ********************************************************************************/

/**
 *	\file	quaternion.h
 *	\author	Adam Cubitt
 *	\brief	defines a quaternion class
 */

#ifndef __QUATERNION__
#define __QUATERNION__


#include <cmath>
#include <iostream>


#include "vector.h"
#include "matrix.h"

class quaternion
{
public:
	///	set x, y, z to 0 and w to 1
	quaternion(){ x = 0.0f; y = 0.0f; z = 0.0f; w = 1.0f; }

	quaternion(const float x_, const float y_, const float z_, const float w_) { x = x_; y = y_; z = z_; w = w_; }
	quaternion(const quaternion &q) { x = q.x; y = q.y; z = q.z; w = q.w; }
	quaternion(const vector3f &v, const float w_) { x = v.x; y = v.y; z = v.z; w = w_; }
	quaternion(const float q[4]){ x = q[0]; y = q[1]; z = q[2]; w = q[3]; }
	~quaternion(){}

	/// assignment to quaternion
	quaternion & operator=(const quaternion& q) { x = q.x; y = q.y; z = q.z; w = q.w; return *this; }

	/// addition of quaternions
	quaternion operator+(const quaternion& q) const { return quaternion(x+q.x, y+q.y, z+q.z, w+q.w); }

	/// subtraction of quaternions
	quaternion operator-(const quaternion& q) const { return quaternion(x-q.x, y-q.y, z-q.z, w-q.w); }

	/// multiplication of quaternions
	quaternion operator*(const quaternion& q) const;

	/// multiplication by value
	quaternion operator*(const float& f) const { return quaternion(x*f, y*f, z*f, w*f); }

	/// division by value
	quaternion operator/(const float& f) const { return quaternion(x/f, y/f, z/f, w/f); }

	/// comparing quaternion equality
	bool operator==(const quaternion& q) const { return FCOMPARE(x,q.x)&&FCOMPARE(y,q.y)&&FCOMPARE(z,q.z)&&FCOMPARE(w,q.w); }

	/// comparing quaternion inequality
	bool operator!=(const quaternion& q) const { return !(FCOMPARE(x,q.x)&&FCOMPARE(y,q.y)&&FCOMPARE(z,q.z)&&FCOMPARE(w,q.w)); }

	/// returns the quaternion as a euler rotation in rads
	vector3f asEulerRad() const;
	/// returns the quaternion as a euler rotation in degrees
	vector3f asEulerDeg() const;
	/// sets the quaternion to a euler rotation in rads
	void setFromEulerRad(const vector3f &v);
	/// Sets the quaternion to a euler rotation in degs
	void setFromEulerDeg(const vector3f &v);
	/// Sets the quaternion from the specified matrix
	void setFromMatrix(const matrix &m);

	/// Get the magnitude of the quaternion
	float queryMagnitude() const { return sqrt(x*x+y*y+z*z+w*w); }

	/// Get the inverse quaternion
	quaternion getInverse() const { float sqmag=x*x+y*y+z*z+w*w; return (sqmag)?quaternion(-x, -y, -z, w)/sqmag:*this; }

	/// Get the inverse quaternion assuming it to be normailsed
	quaternion getInverseDirty() const { return quaternion(-x, -y, -z, w); }

	///	Get the normalised quaternion
	quaternion getNormalised() const { float m=queryMagnitude(); return m?*this/m:*this; }

	/// normalise the quaternion
	void normalise() { float m=queryMagnitude(); if(m!=0) *this=*this/m; }

	/// invert the quaternion
	void invert() { float sqmag=x*x+y*y+z*z+w*w; x=-x/sqmag; y=-y/sqmag; z=-z/sqmag; w=w/sqmag; }

	/// invert the quaternion assuming it to be noramlised
	void invertDirty() { x=-x; y=-y; z=-z; }

	/// assign this quaternion to a matrix
	void toMatrix(matrix &T) const;

	/// get the quaternion as a matrix
	matrix getAsMatrix() const;

	/// Defines the >> operator as a friend, can be used to load quaternion data
	friend std::istream& operator >> (std::istream& is, quaternion& q) { return is >> q.x >> q.y >> q.z >> q.w; }

	/// Defines the << operator as a friend, can be used to output quaternion data
	friend std::ostream& operator << (std::ostream& os, quaternion& q) { return os << q.x << " " << q.y << " " << q.z << " " << q.w; }

	/// spherically interpolate quaternions
	static quaternion sphericallyInterpolate(const quaternion& q1, const quaternion& q2, const float& t);

	/// the quaternion data
	union {
		struct {
			float x;
			float y;
			float z;
			float w;
		};
		float data[4];
	};
};


#endif

