/********************************************************************************
 *
 *  urbanVictory engine
 *  Copyright (C) 2003, 2004 Adam Cubitt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *										
 ********************************************************************************/


/**
 *	\file	vector.h
 *	\author	Adam Cubitt
 *	\brief	defines vector classes
 */




#ifndef __VECTOR_H__
#define __VECTOR_H__




#include <iostream>
#include <cmath>

#include "utilities.h"



/// 3d vector class
class vector3f
{
public:
	vector3f() { x=0; y=0; z=0; }
	vector3f(const float x_, const float y_, const float z_) { x=x_; y=y_; z=z_; }
	vector3f(const float v[3] ) { x=v[0]; y=v[1]; z=v[2]; }
	vector3f(const vector3f& v) { x=v.x; y=v.y; z=v.z; }
	
	~vector3f() {}

	///	assignment from vector
	vector3f & operator=(const vector3f& v) { x = v.x; y = v.y; z = v.z; return *this; }

	///	assignment from array
	vector3f & operator=(const float v[3]) { x = v[0]; y = v[1]; z = v[2]; return *this; }

	/// addition of vectors
	vector3f operator+(const vector3f& v) const { return vector3f(x+v.x, y+v.y, z+v.z); }
	
	/// addition of vectors
	vector3f operator+=(const vector3f& v) { return vector3f(x+=v.x, y+=v.y, z+=v.z); }

	/// subtraction of vectors
	vector3f operator-() const { return vector3f(-x, -y, -z); }

	/// subtraction of vectors
	vector3f operator-(const vector3f& v) const { return vector3f(x-v.x, y-v.y, z-v.z); }

	/// subtraction of vectors
	vector3f operator-=(const vector3f& v) { return vector3f(x-=v.x, y-=v.y, z-=v.z); }

	/// multiplication of vectors by a value
	vector3f operator*(const float& f) const { return vector3f(x*f, y*f, z*f); }
	friend vector3f operator*(const float &f, const vector3f &v) { return v*f; }
	
	/// multiplication of vectors by a value
	vector3f operator*=(const float& f) { return vector3f(x*=f, y*=f, z*=f); }

	/// division of vectors by a value
	vector3f operator/(const float& f) const { return vector3f(x/f, y/f, z/f); }

	/// division of each element in a vector by each element in another vector
	vector3f operator/(const vector3f& v) const { return vector3f(x/v.x, y/v.y, z/v.z); }
	
	/// division of vectors by a value
	vector3f operator/=(const float& f) { return vector3f(x/=f, y/=f, z/=f); }

	/// equality operator for vectors
	bool operator==(const vector3f& v) const { return FCOMPARE(x,v.x)&&FCOMPARE(y,v.y)&&FCOMPARE(z,v.z); }

	/// inequality operator for vectors
	bool operator!=(const vector3f& v) const { return !(FCOMPARE(x,v.x)&&FCOMPARE(y,v.y)&&FCOMPARE(z,v.z)); }
	
	/// Gets the magnitude of the vector
	float mag() const {	return sqrt(x*x+y*y+z*z); }
	float magSquared() const { return x*x+y*y+z*z; }
	bool zeroLength() const { return FCOMPARE(x,0.0f) && FCOMPARE(y,0.0f) && FCOMPARE(z,0.0f); }

	/// Gets a unit vector in the direction of this vector
	vector3f getNormalisedVector() const { float m = mag(); return m?*this/m:*this; }

	/// Normalises this vector (sets it to unit vector)
	vector3f normalise() { float m = mag(); if(m!=0) *this = *this/m; return *this; }

	bool normalised() { return magSquared()==1; }

	/// gets rotation vector from direction vector
	vector3f getAsRotationVector();

	/// Dot product
	float dot(const vector3f& v) const { return (x*v.x + y*v.y + z*v.z); }
	
	///	Cross product
	vector3f cross(const vector3f& v) const { return vector3f(y*v.z-v.y*z, z*v.x-v.z*x, x*v.y-v.x*y); }

	///	Declares >> operator as a friend, can be used to load vector data
	friend std::istream &operator >> (std::istream& is, vector3f& v) { return is >> v.x >> v.y >> v.z; }

	/// Declares << operator as a friend, can be used to ouput vector data
	friend std::ostream &operator << (std::ostream& os, vector3f& v) { return os << "[" << v.x << " " << v.y << " " << v.z << "]"; }

	/// linear interpolation of vectors
	static vector3f linearlyInterpolate(const vector3f& v1, const vector3f& v2, const float &t)
	{
		return vector3f(	v1.x+(v2.x-v1.x)*t,
							v1.y+(v2.y-v1.y)*t,
							v1.z+(v2.z-v1.z)*t);
	}

	/// The vector data
	union {
		struct {
			float x;
			float y;
			float z;
		};
		struct {
			float r;
			float g;
			float b;
		};
		struct { // y is up, z is depth/forward
			float pitch;	//rotation about x
			float yaw;		// rotation about y
			float roll;		// rotation about z
		};
		float data[3];
	};
};

typedef vector3f v3f;




class vector2f
{
public:
	vector2f() { x=0; y=0; }
	vector2f(float x_, float y_) { x=x_; y=y_; }
	vector2f(float v[2] ) { x=v[0]; y=v[1]; }
	vector2f(const vector2f& v) { x=v.x; y=v.y; }

	~vector2f() {}

	///	assignment from vector
	vector2f & operator=(const vector2f& v) { x = v.x; y = v.y; return *this; }

	///	assignment from array
	vector2f & operator=(const float v[3]) { x = v[0]; y = v[1]; return *this; }

	/// addition of vectors
	vector2f operator+(const vector2f& v) const { return vector2f(x+v.x, y+v.y); }

	/// addition of vectors
	vector2f operator+=(const vector2f& v) { return vector2f(x+=v.x, y+=v.y); }

	/// subtraction of vectors
	vector2f operator-() const { return vector2f(-x, -y); }

	/// subtraction of vectors
	vector2f operator-(const vector2f& v) const { return vector2f(x-v.x, y-v.y); }

	/// subtraction of vectors
	vector2f operator-=(const vector2f& v) { return vector2f(x-=v.x, y-=v.y); }

	/// multiplication of vectors by a value
	vector2f operator*(const float& f) const { return vector2f(x*f, y*f); }
	friend vector2f operator*(const float &f, const vector2f &v) { return v*f; }

	/// multiplication of vectors by a value
	vector2f operator*=(const float& f) { return vector2f(x*=f, y*=f); }

	/// division of vectors by a value
	vector2f operator/(const float& f) const { return vector2f(x/f, y/f); }

	/// division of each element in a vector by each element in another vector
	vector2f operator/(const vector2f& v) const { return vector2f(x/v.x, y/v.y); }

	/// division of vectors by a value
	vector2f operator/=(const float& f) { return vector2f(x/=f, y/=f); }

	/// equality operator for vectors
	bool operator==(const vector2f& v) const { return FCOMPARE(x,v.x)&&FCOMPARE(y,v.y); }

	/// inequality operator for vectors
	bool operator!=(const vector2f& v) const { return !FCOMPARE(x,v.x)||!FCOMPARE(y,v.y); }

	/// Gets the magnitude of the vector
	float mag() const {	return sqrt(x*x+y*y); }

	/// Gets a unit vector in the direction of this vector
	vector2f getNormalisedVector() const { float m = mag(); return m?*this/m:*this; }

	/// Normalises this vector (sets it to unit vector)
	vector2f normalise() { float m = mag(); if(m!=0) *this = *this/m; return *this; }

	/// Dot product
	float dot(const vector2f& v) const { return (x*v.x + y*v.y); }

	///	Declares >> operator as a friend, can be used to load vector data
	friend std::istream &operator >> (std::istream& is, vector2f& v) { return is >> v.x >> v.y; }

	/// Declares << operator as a friend, can be used to ouput vector data
	friend std::ostream &operator << (std::ostream& os, vector2f& v) { return os << "[" << v.x << " " << v.y << "]"; }

	/// linear interpolation of vectors
	static vector2f linearlyInterpolate(const vector2f& v1, const vector2f& v2, const float &t)
	{
		return vector2f(	v1.x+(v2.x-v1.x)*t,
							v1.y+(v2.y-v1.y)*t	);
	}

	/// The vector data
	union {
		struct {
			float x;
			float y;
		};
		struct {
			float u;
			float v;
		};
		float data[2];
	};
};

typedef vector2f v2f;





class vector2i
{
public:
	vector2i() { x=0; y=0; }
	vector2i(int x_, int y_) { x=x_; y=y_; }
	vector2i(int v[2] ) { x=v[0]; y=v[1]; }
	vector2i(const vector2i& v) { x=v.x; y=v.y; }

	~vector2i() {}

	///	assignment from vector
	vector2i & operator=(const vector2i& v) { x = v.x; y = v.y; return *this; }

	///	assignment from array
	vector2i & operator=(const int v[3]) { x = v[0]; y = v[1]; return *this; }

	/// addition of vectors
	vector2i operator+(const vector2i& v) const { return vector2i(x+v.x, y+v.y); }

	/// addition of vectors
	vector2i operator+=(const vector2i& v) { return vector2i(x+=v.x, y+=v.y); }

	/// subtraction of vectors
	vector2i operator-() const { return vector2i(-x, -y); }

	/// subtraction of vectors
	vector2i operator-(const vector2i& v) const { return vector2i(x-v.x, y-v.y); }

	/// subtraction of vectors
	vector2i operator-=(const vector2i& v) { return vector2i(x-=v.x, y-=v.y); }

	/// multiplication of vectors by a value
	vector2i operator*(const int& f) const { return vector2i(x*f, y*f); }
	friend vector2i operator*(const int &f, const vector2i &v) { return v*f; }

	/// multiplication of vectors by a value
	vector2i operator*=(const int& f) { return vector2i(x*=f, y*=f); }

	/// division of vectors by a value
	vector2i operator/(const int& f) const { return vector2i(x/f, y/f); }

	/// division of each element in a vector by each element in another vector
	vector2i operator/(const vector2i& v) const { return vector2i(x/v.x, y/v.y); }

	/// division of vectors by a value
	vector2i operator/=(const int& f) { return vector2i(x/=f, y/=f); }

	/// equality operator for vectors
	bool operator==(const vector2i& v) const { return x==v.x && y==v.y; }

	/// inequality operator for vectors
	bool operator!=(const vector2i& v) const { return x!=v.x || y!=v.y; }

	/// Gets the magnitude of the vector
	float mag() const {	return sqrt((float)x*x+y*y); }

	/// Dot product
	int dot(const vector2i& v) const { return (x*v.x + y*v.y); }

	///	Declares >> operator as a friend, can be used to load vector data
	friend std::istream &operator >> (std::istream& is, vector2i& v) { return is >> v.x >> v.y; }

	/// Declares << operator as a friend, can be used to ouput vector data
	friend std::ostream &operator << (std::ostream& os, vector2i& v) { return os << "[" << v.x << " " << v.y << "]"; }

	/// The vector data
	union {
		struct {
			int x;
			int y;
		};
		int data[2];
	};
};

typedef vector2i v2i;





#endif

