/********************************************************************************
 *
 *  urbanVictory engine
 *  Copyright (C) 2003, 2004 Adam Cubitt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *										
 ********************************************************************************/

/**
 *	\file	id.h
 *	\author	Adam Cubitt
 *	\brief	defines id related classes
 */


#ifndef __ID__
#define __ID__


#include <list>
#include <cassert>
#include <iostream>




typedef unsigned int uInt;
#define Assert(a, b) assert( a && b)


class noRecycling;
template<class>
class wrapAround;

/// storage class for ids
/**
 *	ids consist of a number formed of two parts
 *	an index and a magic number
 *	\arg bitsIndex: the number of bits to use for the index
 *	\arg bitsMagic: the number of bits to use for the magic number
*/
template<uInt bitsIndex, uInt bitsMagic>
class id
{
public:
	id<bitsIndex,bitsMagic>() : m_id(0) {}
	id<bitsIndex,bitsMagic>(const uInt& v) { m_id = v; }
	~id<bitsIndex,bitsMagic>() {}

	/// assignment from an id of the same type
	id<bitsIndex,bitsMagic> & operator=(const id<bitsIndex,bitsMagic>& id_) { m_id = id_; return *this; }
	/// assignment from a uInt
	id<bitsIndex,bitsMagic> & operator=(const uInt& v) { m_id = v; return *this; }
	/// uInt conversion operator
	operator uInt () const { return m_id; }
	/// equality comparison operator for an id of the same type
	bool operator==(const id<bitsIndex,bitsMagic>& id_) const { return m_id == id_; }
	/// inequality comparison operator for an id of the same type
	bool operator!=(const id<bitsIndex,bitsMagic>& id_) const { return m_id != id_; }
	/// returns true if id is null
	bool isNull() const { return m_id == 0; }
	/// assignment from an id of a different type
	template<uInt bitsIndex_, uInt bitsMagic_>
	id<bitsIndex,bitsMagic> & operator=(const id<bitsIndex_,bitsMagic_>& id_)
	{
		Assert(id_.m_index<=MAX_INDEX && id_.m_magic<=MAX_MAGIC, "invalid assignment to an id from an id of a different type");
		m_index = id_.m_index;
		m_magic = id_.m_magic;
		return *this;
	}

	/// contains info about the id type
	enum
	{
		BITS_INDEX = bitsIndex,
		BITS_MAGIC = bitsMagic,
		MAX_INDEX = (1<<bitsIndex) - 1,
		MAX_MAGIC = (1<<bitsMagic) - 1
	};

	/// contains the id data
	union
	{
		struct
		{
			uInt m_index : bitsIndex;
			uInt m_magic : bitsMagic;
		};

		uInt m_id : (bitsIndex+bitsMagic);
	};
protected:
private:
};



/// stores the number used to generate new unique ids
/**
 *	(templated with the maximum id size in bits)
 */
template< uInt numBits >
class newIdGen
{
protected:
	newIdGen() : m_idGen(0) {}

	/// stores the number of bits per id and the
	/// resulting maximum id possible
	enum
	{
		NUM_BITS = numBits,
		MAX_ID = ( 1 << numBits ) - 1
	};

	/// stores the last new id generated
	uInt m_idGen : numBits;
};


/// class for generating ids
/**
 *	\arg recyclingPolicy:	defines how ids that have been released are reused
 *	 						(default is noRecycling)
 *	\arg wrapAroundPolicy:	defines what happens when the maximum possible id
 *	 						is exceeded (defualt is wrapAround)
 *	\arg numBits:	the size of the generated ids in bits (default is 32)
 */
template<
	class recyclingPolicy = noRecycling,
	template <class> class wrapAroundPolicy = wrapAround,
	uInt numBits = 32
>
class idGen
	: public recyclingPolicy
	, public wrapAroundPolicy< newIdGen<numBits> >
{
public:
	/// gets an id
	uInt getId()
	{
		// attempt to get a recycled id (dependant on the policiy used)
		uInt id = getRecycledId();

		// if we didn't get a recycled id then get a new one
		if(!id) id = getNewId();

		// otherwise attempt to generate a new id and return the result
		return id;
	}


protected:
private:
};




/// recyclingPolicy: released ids are not recycled
class noRecycling
{
public:
	/// does nothing
	void releaseId(uInt id) {}

protected:
	/// always returns null
	uInt getRecycledId() { return 0; }

private:
};


/// recyclingPolicy: released ids are stored in a cache and reused
class recycleReleasedIds
{
public:
	/// adds the id to the cache of reusable ids
	void releaseId(uInt id)
	{
		m_releasedIds.push_back(id);
	}

protected:
	/// returns the first id in the cache of reusable ids
	/// or null if there are none
	uInt getRecycledId()
	{
		if(m_releasedIds.size())
		{
			uInt id = *(m_releasedIds.begin());
			m_releasedIds.pop_front();
			return id;
		} else
		{
			return 0;
		}
	}

private:
	/// the cache of reusable ids
	std::list<uInt> m_releasedIds;
};


/// wrapAroundPolicy: swhen the maximum id is reached id generation restarts from 1
template<class _newIdGen>
class wrapAround : public _newIdGen
{
public:
	/// return the next new id (1 if the maximum id has been reached)
	uInt getNewId()
	{
		if(++m_idGen==0)
		{
			m_idGen = 1;
		}
		return m_idGen;
	}
protected:
private:
	/// the last new id generated
	//uInt m_idGen;
};


/// wrapAroundPolicy:
/// when the maximum id is reached the generation function asserts
/// with the message "exceeded maximum id"
template<class _newIdGen>
class assertOnWrapAround : public _newIdGen
{
public:
	/// return the next new id (or assert if the maximum id has been
	/// reached)
	uInt getNewId()
	{
		++m_idGen;
		Assert(m_idGen,"exceeded maximum id");
		return m_idGen;
	}
protected:
private:
};




/// wrapAroundPolicy: 
/// when the maximum id is exceeded the null id (0) is returned
template<class _newIdGen>
class noWrapAround : public _newIdGen
{
public:
	/// return the next new id (or the null id if the maximum id has
	/// been reached)
	uInt getNewId()
	{
		// if this is the maximum id
		if(m_idGen==MAX_ID)
		{
			// return the null id
			return 0;
		}
		return ++m_idGen;
	}
protected:
private:
};

#endif
