
#include "virtualMachine.h"




process::process(uInt maxThreads, processId pid, program& prog)	: m_pid(pid)
																, m_program(prog)
																, m_threadCnt(1)
																, m_terminated(false)
{
	ASSERT(maxThreads!=0, "cannot create a process with no threads");

	resize(maxThreads);
	begin()->spawn(0);
}


threadId process::spawnThread(address addr)
{
	iterator it = begin();
	for(uInt tid=0; it!=end(); ++tid, ++it)
	{
		if(!it->m_inUse)
		{
			it->spawn(addr);
			m_threadCnt++;

			return tid+1;
		}
	}

	return 0;
}


void process::killThread(threadId tid)
{
	if(tid>0 && tid<=size() && (*this)[tid-1].m_inUse)
	{
		(*this)[tid-1].kill();
		m_threadCnt--;
	}
}


void process::step(instructionSet& instrSet)
{
	iterator it = begin();
	for( ; it!=end(); ++it)
	{
		if(it->m_inUse)
		{
			it->m_PC%=m_program.size();
			instrSet[m_program[it->m_PC]%instrSet.size()](*this, it->m_PC);
			it->m_PC%=m_program.size();
		}
	}
}


virtualMachine::~virtualMachine()
{
	clean();
}

void virtualMachine::clean()
{
	std::vector<process*>::iterator it=m_pProcesses.begin();
	for( ;it!=m_pProcesses.end();++it)
	{
		if(*it) delete (*it);
	}
}


instruction virtualMachine::addInstruction(instructionPtr instrPtr)
{
	push_back(instrPtr);
	return static_cast<instruction>(size());
}


bool virtualMachine::replaceInstruction(instruction instr, instructionPtr instrPtr)
{
	if(instr>=size())
	{
		WARN("instruction does not exist");
		return false;
	}

	if(!instrPtr)
	{
		WARN("NULL pointer when replacing instruction");
		return false;
	}

	(*this)[instr] = instrPtr;

	return true;
}


processId virtualMachine::run(program& prog)
{
	if(size()==0)
	{
		WARN("cannot execute programs on virtual machine with no instruction set");
		return 0;
	}

	if(prog.size()==0)
	{
		WARN("cannot run 0 length program");
		return 0;
	}

	uInt pid = m_processIdGen.getId();
	if(pid)
	{
		if(pid>m_pProcesses.size())
		{
			m_pProcesses.resize(pid, NULL);
		}

		m_pProcesses[pid-1] = new process(m_maxThreadsPerProcess, pid, prog);
		m_numProcesses++;

		return pid;
	}

	return 0;
}


void virtualMachine::kill(processId pid)
{
	if(pid == 0 || pid > m_pProcesses.size() || !m_pProcesses[pid-1]) return;

	m_processIdGen.releaseId(pid);
	delete m_pProcesses[pid-1];
	m_pProcesses[pid-1] = NULL;

	m_numProcesses--;
}


process* virtualMachine::getProcess(processId pid)
{
	if(pid == 0 || pid > m_pProcesses.size() || !m_pProcesses[pid-1]) return NULL;
	return m_pProcesses[pid-1];
}


void virtualMachine::step()
{
	std::vector<process*>::iterator pit = m_pProcesses.begin();
	while(pit!=m_pProcesses.end())
	{
		if(*pit)
		{
			(*pit)->step(*this);

			if((*pit)->isTerminated())
			{
				m_processIdGen.releaseId((*pit)->queryId());
				delete (*pit);
				*pit = NULL;
			}
		}
		++pit;
	}
}


void virtualMachine::setMaxThreadsPerProcess(uInt numThreads)
{
	if(numThreads==0)
	{
		WARN("maximum number of threads must be larger than 0");
		return;
	}

	m_maxThreadsPerProcess = numThreads;
}


void virtualMachine::setNumUsedCycles(float num)
{
	m_usedCycles = num;
	if(m_usedCycles>1.0f) m_usedCycles = 1.0f;
}


void virtualMachine::update(float deltaT)
{
	if(m_idleCnt >= 1.0f/m_usedCycles)
	{
		void step();
	} else
	{
		m_idleCnt++;
	}
}

