
#include "virtualMachine.h"




process::process(program& prog, processId pid)	: m_pProgram(&prog)
												, m_id(pid)
												, m_threadCnt(1)
{
	push_back(new thread(0));
}


process::~process()
{
	clean();
}


void process::clean() 
{ 
	iterator tit = begin();
	for( ; tit!=end();++tit)
	{
		if(*tit) delete (*tit);
	}

	resize(0);

	m_threadCnt = m_id = 0; 
	m_pProgram = NULL;
}


program& process::getProgram() 
{ 
	ASSERT(m_pProgram, "cannot manipulate terminated process");
	return *m_pProgram; 
}


threadId process::spawnThread(address addr)
{
	iterator it = begin();
	for(uInt tid=0; it!=end(); ++tid, ++it)
	{
		if((*it)->isTerminated())
		{
			(*it)->spawn(addr);

			return tid+1;
		}
	}

	push_back(new thread(addr));
	m_threadCnt++;

	return static_cast<uInt>(size());
}


void process::killThread(threadId tid)
{
	if(tid>0 && tid<=size() && !(*this)[tid-1]->m_terminated)
	{
		(*this)[tid-1]->m_terminated=true;
		m_threadCnt--;
	}
}


bool process::isTerminated() 
{
	return		m_threadCnt==0 
			||	m_pProgram==NULL 
			||	m_pProgram->size()==0;
}



virtualMachine::~virtualMachine() 
{
	std::vector<process*>::iterator pit = m_pProcesses.begin();
	for( ; pit!=m_pProcesses.end();++pit)
	{
		if(*pit) delete (*pit);
	}
}


instruction virtualMachine::addInstruction(instructionPtr instrPtr)
{
	push_back(instrPtr);
	return static_cast<instruction>(size());
}


bool virtualMachine::replaceInstruction(instruction instr, instructionPtr instrPtr)
{
	if(instr>=size())
	{
		WARN("instruction does not exist");
		return false;
	}

	if(!instrPtr)
	{
		WARN("NULL pointer when replacing instruction");
		return false;
	}

	(*this)[instr] = instrPtr;

	return true;
}


processId virtualMachine::run(program& prog)
{
	if(size()==0)
	{
		WARN("cannot execute programs on virtual machine with no instruction set");
		return 0;
	}

	uInt pid = m_processIdGen.getId();
	if(pid)
	{
		if(pid>m_pProcesses.size())
		{
			process* pNewProc = new process(prog, pid);
			m_pProcesses.push_back(pNewProc);
		} else
		{
			process* pNewProc = m_pProcesses[pid-1];
			pNewProc->m_pProgram = &prog;
			pNewProc->spawnThread(0);
		}

		m_numProcesses++;
		return pid;
	}

	return 0;
}


void virtualMachine::kill(processId pid)
{
	if(		pid > 0 
		&&	pid <= m_pProcesses.size() 
		&&	!m_pProcesses[pid-1]->isTerminated()	) 
	{
		kill(*(m_pProcesses[pid-1]));
	}
}


process* virtualMachine::getProcess(processId pid)
{
	if(		pid > 0 
		&&	pid <= m_pProcesses.size() 
		&&	!m_pProcesses[pid-1]->isTerminated()	) 
	{
		return m_pProcesses[pid-1];
	}

	return NULL;
}


void virtualMachine::step()
{
	m_activeProcesses = false;

	for( processId pid=0; pid<m_pProcesses.size(); ++pid)
	{
		process& proc = *(m_pProcesses[pid]);

		for(threadId tid=0; tid<proc.size();++tid)
		{
			if(canBeExecuted(proc))
			{
				m_activeProcesses = true;

				thread& t = *(proc[tid]);

				if(!t.isTerminated())
				{
					instruction instr = proc.m_pProgram->read(t)%static_cast<uInt>(size());
					(*this)[instr](*this,proc,t);
				}

			} else if(shouldBeTerminated(proc))
			{
				kill(proc);
			}
		}

	}
}


void virtualMachine::setNumUsedCycles(float num)
{
	m_usedCycles = num;
	if(m_usedCycles>1.0f) m_usedCycles = 1.0f;
}


void virtualMachine::update(float deltaT)
{
	if(m_usedCycles<=0.0f) return;

	if(m_idleCnt >= 1.0f/m_usedCycles)
	{
		step();
	} else
	{
		m_idleCnt++;
	}
}


bool virtualMachine::canBeExecuted(process& proc)
{
	return !proc.isTerminated();
}


bool virtualMachine::shouldBeTerminated(process& proc)
{
	return proc.m_id!=0;
}


void virtualMachine::kill(process& proc)
{
	m_processIdGen.releaseId(proc.m_id);
	proc.clean();
	m_numProcesses--;
}


