

#ifndef __virtualMachine_h__
#define __virtualMachine_h__


#include "common.h"
#include "id.h"
#include "program.h"



/// holds the data for an individual thread
class thread
{
	friend class process;

public:
	/// initialise to terminated state
	thread()	: m_PC(0)
				, m_terminated(true) {}

	/// initialise to specified address
	thread(address PC)	: m_PC(PC)
						, m_terminated(false) {}

	~thread() {}

	/// set the thread to the specified address
	/// and flag as active
	void spawn(const address& PC) { m_PC = PC; m_terminated = false; }

	/// get the current address of the thread
	address getAddress() { return m_PC; }

	/// test if the thread is terminated
	bool isTerminated() { return m_terminated; }

	/// move the thread to the next address
	thread& operator++() { ++m_PC; return *this; }

	/// set the address to the specified address
	thread& operator=(const address& addr) { m_PC = addr; return *this; }

	/// implicit conversion to the current address
	operator address() { return m_PC; }

protected:
	/// the current address of the thread
	address m_PC;

	/// whether the thread is terminated or active
	bool m_terminated;

private:
};


/// holds the data for a process
/*
 *	a process is a collection of one or more threads
 */
class process
	: protected std::vector<thread*>
{
	friend class virtualMachine;
	friend class std::vector<thread*>;

public:
	typedef void(*instructionPtr)(virtualMachine&, process&, thread&);
	typedef std::vector<instructionPtr> instructionSet;

	/// initialise active process
	/*
	 *	creates a thread at address 0 of the program
	 */
	process(program& prog, uInt pid);
	~process();

	/// terminates and cleans up the thread
	void clean();

	/// get the program being run
	program& getProgram();

	/// get the id of the process
	processId getId() { return m_id; }

	/// spawn a new thread at the specified address
	threadId spawnThread(address addr);

	/// kill the specified thread
	void killThread(threadId tid);

	/// get the number of threads currently being executed
	uInt getNumThreads() { return m_threadCnt; }	
	
	/// test if the process has terminated
	/*
	 *	a thread is considered terminated if it
	 *	has no threads, has no program, or the
	 *	program size is 0
	 */
	virtual bool isTerminated();

protected:
	/// initalise dead process
	process() : m_pProgram(NULL), m_id(0) {}

	/// pointer to the program being executed
	program* m_pProgram;

	/// the number of active threads
	uInt m_threadCnt;

	/// the id of the process
	processId m_id;
	
private:
};


typedef void(*instructionPtr)(virtualMachine&, process&, thread&);
typedef std::vector<instructionPtr> instructionSet;



/// a basic virtual machine with multi-threaded proccesses
class virtualMachine
	: protected instructionSet
{
public:
	virtualMachine()	: m_usedCycles(1.0f)
						, m_maxThreadsPerProcess(1)
						, m_idleCnt(0)
						, m_numProcesses(0)
						, m_activeProcesses(false) {}

	~virtualMachine();

	/// add an instruction
	instruction addInstruction(instructionPtr instrPtr);

	/// replace the specified instruction with a instruction
	bool replaceInstruction(instruction instr, instructionPtr instrPtr);

	/// get the current number of instructions
	uInt getNumInstructions() { return static_cast<uInt>(size()); }

	/// run the specified program and return the id of the process
	processId run(program& prog);

	/// kill the specified process
	void kill(processId pid);

	/// get a pointer to the specified process
	process* getProcess(processId pid);

	/// get the number of processes being executed
	uInt getNumProcesses() { return m_numProcesses; }

	/// set the maximum number of threads per process
	void setMaxThreadsPerProcess(uInt n) { m_maxThreadsPerProcess=n; }

	/// get the maximum number of threads per process
	uInt getMaxThreadsPerProcess() { return m_maxThreadsPerProcess; }

	/// set the number of steps per update
	/*
	 *	\note should be between 0 and 1
	 */
	void setNumUsedCycles(float num);

	/// test if there are still active processes
	bool activeProcesses() { return m_activeProcesses; }
	
	/// step the virtual machine
	/*
	 *	executes one instruction for each process
	 */
	void step();

	/// update the virtual machine
	void update(float deltaT);

protected:
	/// test if the specified thread can be executed
	virtual bool canBeExecuted(process& proc);

	/// test if the process should be terminated
	/*
	 *	only called if canBeExecuted returns false so
	 *	the condition of canBeExecuted can be assumed
	 */
	virtual bool shouldBeTerminated(process& proc);

	/// kills the specified process
	virtual void kill(process& proc);

	/// vector of allocated processes
	std::vector<process*> m_pProcesses;

	/// number of running processes
	uInt m_numProcesses;

	/// max number of threads per process
	uInt m_maxThreadsPerProcess;

	/// used to generate proccess ids and keep
	/// track of terminated processes
	idGen<recycleReleasedIds, noWrapAround> m_processIdGen;

	float m_usedCycles;
	uInt m_idleCnt;

	/// whether there were active processes in the
	/// last step
	bool m_activeProcesses;

private:
};



#endif

