
#include "virtualMachine.h"



virtualMachine::~virtualMachine() 
{
	// delete all processes
	std::vector<process*>::iterator pit = m_pProcesses.begin();
	for( ; pit!=m_pProcesses.end();++pit)
	{
		if(*pit) delete (*pit);
	}
}


instruction virtualMachine::addInstruction(instructionPtr instrPtr)
{
	push_back(instrPtr);
	return static_cast<instruction>(size());
}


bool virtualMachine::replaceInstruction(instruction instr, instructionPtr instrPtr)
{
	if(instr>=size())
	{
		WARN("instruction does not exist");
		return false;
	}

	if(!instrPtr)
	{
		WARN("NULL pointer when replacing instruction");
		return false;
	}

	(*this)[instr] = instrPtr;

	return true;
}


processId virtualMachine::run(program& prog)
{
	if(size()==0)
	{
		WARN("cannot execute programs on virtual machine with no instruction set");
		return 0;
	}

	// get a process id
	uInt pid = m_processIdGen.getId();
	if(pid)
	{
		// create the process
		if(pid>m_pProcesses.size())
		{
			// new process
			process* pNewProc = new process(prog, pid);
			m_pProcesses.push_back(pNewProc);
		} else
		{
			// recycled process
			process* pNewProc = m_pProcesses[pid-1];
			pNewProc->m_pProgram = &prog;
			pNewProc->spawnThread(0);
		}

		m_numProcesses++;
		return pid;
	}

	return 0;
}


void virtualMachine::kill(processId pid)
{
	if(		pid > 0 
		&&	pid <= m_pProcesses.size() 
		&&	!m_pProcesses[pid-1]->isTerminated()	) 
	{
		kill(*(m_pProcesses[pid-1]));
	}
}


process* virtualMachine::getProcess(processId pid)
{
	if(		pid > 0 
		&&	pid <= m_pProcesses.size() 
		&&	!m_pProcesses[pid-1]->isTerminated()	) 
	{
		return m_pProcesses[pid-1];
	}

	return NULL;
}


void virtualMachine::step()
{
	m_activeProcesses = false;

	// loop through all processes
	for( processId pid=0; pid<m_pProcesses.size(); ++pid)
	{
		process& proc = *(m_pProcesses[pid]);

		// loop through each thread of the process
		for(threadId tid=0; tid<proc.size();++tid)
		{
			// check if the process can still be executed
			// (any thread could kill the process so we
			// need to check after each thread has been 
			// updated)
			if(canBeExecuted(proc))
			{
				m_activeProcesses = true;

				thread& t = *(proc[tid]);

				// check if the thread is executing
				if(!t.isTerminated())
				{
					// execute the thread
					t%=static_cast<uInt>(proc.m_pProgram->size());
					instruction instr = proc.m_pProgram->read<instruction>(t)%static_cast<uInt>(size());
					(*this)[instr](*this,proc,t);
				}

			} else if(shouldBeTerminated(proc))
			{
				// if the process is not being executed
				// check if it should be terminated and
				// if so terminate it
				kill(proc);
			}
		}

	}
}


void virtualMachine::setNumUsedCycles(float num)
{
	m_usedCycles = num;
	if(m_usedCycles>1.0f) m_usedCycles = 1.0f;
}


void virtualMachine::update(float deltaT)
{
	if(m_usedCycles<=0.0f) return;

	if(m_idleCnt >= 1.0f/m_usedCycles)
	{
		step();
	} else
	{
		m_idleCnt++;
	}
}


bool virtualMachine::canBeExecuted(process& proc)
{
	return !proc.isTerminated();
}


bool virtualMachine::shouldBeTerminated(process& proc)
{
	return proc.m_id!=0;
}


void virtualMachine::kill(process& proc)
{
	m_processIdGen.releaseId(proc.m_id);
	proc.clean();
	m_numProcesses--;
}


