

#ifndef __virtualMachine_h__
#define __virtualMachine_h__


#include "common.h"
#include "id.h"
#include "process.h"



typedef void(*instructionPtr)(virtualMachine&, process&, thread&);
typedef std::vector<instructionPtr> instructionSet;



/// a basic virtual machine with multi-threaded proccesses
class virtualMachine
	: protected instructionSet
{
public:
	virtualMachine()	: m_usedCycles(1.0f)
						, m_idleCnt(0)
						, m_numProcesses(0)
						, m_activeProcesses(false) {}

	~virtualMachine();

	/// add an instruction
	instruction addInstruction(instructionPtr instrPtr);

	/// replace the specified instruction with a instruction
	bool replaceInstruction(instruction instr, instructionPtr instrPtr);

	/// get the current number of instructions
	uInt getNumInstructions() { return static_cast<uInt>(size()); }

	/// run the specified program and return the id of the process
	processId run(program& prog);

	/// kill the specified process
	void kill(processId pid);

	/// get a pointer to the specified process
	process* getProcess(processId pid);

	/// get the number of processes being executed
	uInt getNumProcesses() { return m_numProcesses; }

	/// set the number of steps per update
	/*
	 *	\note should be between 0 and 1
	 */
	void setNumUsedCycles(float num);

	/// test if there are still active processes
	bool activeProcesses() { return m_activeProcesses; }
	
	/// step the virtual machine
	/*
	 *	executes one instruction for each process
	 */
	void step();

	/// update the virtual machine
	void update(float deltaT);

protected:
	/// test if the specified process can be executed
	virtual bool canBeExecuted(process& proc);

	/// test if the process should be terminated
	/*
	 *	only called if canBeExecuted returns false so
	 *	the condition of canBeExecuted can be assumed
	 */
	virtual bool shouldBeTerminated(process& proc);

	/// kills the specified process
	virtual void kill(process& proc);

	/// vector of allocated processes
	std::vector<process*> m_pProcesses;

	/// number of running processes
	uInt m_numProcesses;

	/// used to generate proccess ids and keep
	/// track of terminated processes
	idGen<recycleReleasedIds, noWrapAround> m_processIdGen;

	float m_usedCycles;
	uInt m_idleCnt;

	/// whether there were active processes during the
	/// last step
	bool m_activeProcesses;

private:
};



#endif

